try:
    import pygmsh
except ImportError:
    print("This example needs the 'pygmsh' package.")
    print("    pip install pygmsh")
    import sys
    sys.exit(77) # make ctest mark test as skipped

import dune.fem, dune.vem
from uzawa import Uzawa # saddle point solver
from dune.ufl import Constant, DirichletBC
from ufl import SpatialCoordinate, TrialFunction, TestFunction,\
                inner, dot, div, nabla_grad, dx, as_vector, sym

dune.fem.threading.use = 8
coarse = True
order  = 3
T      = 0.5 if coarse else 2.5
tau    = T / 8000
mu     = 0.001

# use pygmsh to define domain with cylinder
with pygmsh.occ.Geometry() as geom:
    geom.set_mesh_size_callback( lambda dim, tag, x, y, z, lc:
        min(0.02+1.5*( (x-0.2)**2+(y-0.2)**2), 0.25 if coarse else 0.04)
    )
    rectangle = geom.add_rectangle([0, 0, 0], 2.2, 0.41)
    cylinder = geom.add_disk([0.2, 0.2, 0.0], 0.05)
    geom.boolean_difference(rectangle, cylinder)
    mesh = geom.generate_mesh()
    points, cells = mesh.points, mesh.cells_dict
    # extract dictionary for grid construction
    domain = {
        "vertices": points[:, :2].astype(float),
        "simplices": cells["triangle"].astype(int),
    }
gridView = dune.vem.polyGrid(domain)
print(gridView.size(0),flush=True)

# set up spaces for velocity and pressure
spcU = dune.vem.divFreeSpace( gridView, order=order)
spcP = dune.fem.space.finiteVolume( gridView )
u_h   = spcU.interpolate([0, 0], name="u_h")
u_h_n = spcU.interpolate([0, 0], name="u_h")
p_h   = spcP.interpolate(0, name="p_h")

# set up models for saddle point problem
x            = SpatialCoordinate(spcU)
mu,nu        = Constant(mu, "mu"), Constant(1/tau, "nu")
inflow       = [6 * x[1] * (0.41 - x[1]) / 0.41**2, 0] # 6->18
dbc_u_in     = DirichletBC(spcU, inflow, x[0] <= 1e-8)
dbc_u_noflow = DirichletBC(spcU, [0, 0], None)
dbc_u_out    = DirichletBC(spcU, [None,None], 2.2 - x[0] <= 1e-8)
dbc_p_out    = DirichletBC(spcP, 0, 2.2 - x[0] <= 1e-8)

u,v   = TrialFunction(spcU), TestFunction(spcU)
p,q   = TrialFunction(spcP), TestFunction(spcP)
b         = lambda u1,u2,v: dot( dot(u1, nabla_grad(u2)), v)
epsilon   = lambda w: sym(nabla_grad(w))
a         = lambda u,v: inner(epsilon(u), epsilon(v))
mainModel = ( nu*dot(u-u_h_n,v) + b(u_h_n,u_h_n,v) + 2*mu*a(u,v) ) * dx
gradModel = -p*div(v) * dx
divModel  = -div(u)*q * dx

# use a Uzawa type algorithm to solve the problem
uzawa = Uzawa(gridView, spcU, spcP,
              [dbc_u_in,dbc_u_out,dbc_u_noflow],
              mainModel, gradModel, divModel, mu, nu,
              tolerance=1e-6, precondition=True, verbose=False)

# time loop
if coarse:
    gridView.writeVTK("cylinder_grid", pointdata=[p_h], pointvector=[u_h])
    vtk = gridView.sequencedVTK("cylinder",
                  subsampling=order, pointdata=[p_h], pointvector=[u_h])
else:
    gridView.writeVTK("cylinderFine_grid", pointdata=[p_h],pointvector=[u_h])
    vtk = gridView.sequencedVTK("cylinderFine",
                  subsampling=order, pointdata=[p_h], pointvector=[u_h])
vtk()
time, saveStep = 0, 0.05
nextSave = saveStep
while time < T:
    u_h_n.assign(u_h)
    info = uzawa.solve([u_h,p_h])
    time += tau
    if time > nextSave:
        print(time, tau, info, flush=True)
        # print("  *********** saving ********  ")
        vtk()
        nextSave += saveStep



from ufl import *
import dune.ufl

from scipy.sparse.linalg import spsolve

from script import runTest, checkEOC, interpolate, getParameters

parameters = {"newton.linear.tolerance": 1e-8,
              "newton.tolerance": 5e-6,
              "newton.lineSearch": "simple",
              "newton.linear.verbose": False,
              "newton.verbose": True
              }

mass = dune.ufl.Constant(1, "mu")

def curlfree(space, exact):
    u = TrialFunction(space)
    v = TestFunction(space)

    a = (div(u)*div(v) + dot(u,v) ) * dx
    b = dot( -grad(div(exact)) + exact, v) * dx
    dbc = [dune.ufl.DirichletBC(space, exact, i+1) for i in range(4)]

    df = space.interpolate(exact,name="solution")
    scheme = dune.vem.vemScheme(
                  [a==b, *dbc], space,
                  solver="cg",
                #   solver=("suitesparse","umfpack"),
                  parameters=parameters,
                  gradStabilization=0,
                  massStabilization=mass)
    info = scheme.solve(target=df)

    edf = exact-df
    err = [inner(edf,edf),
           inner(div(edf),div(edf))]

    return err

def runTestCurlfree(order):
    x = SpatialCoordinate(triangle)
    ln, lm, Lx, Ly = getParameters()
    # print(ln, lm, Lx, Ly)
    exact = -grad( (cos(ln/Lx*pi*x[0])*cos(lm/Ly*pi*x[1])) )
    spaceConstructor = lambda grid, r: dune.vem.curlFreeSpace( grid,
                                                               order=order )

    expected_eoc = [order+1,order+1]
    if (interpolate()):
      eoc = runTest(exact, spaceConstructor, interpolate_secondorder)
    else:
      eoc = runTest(exact, spaceConstructor, curlfree)

    return eoc, expected_eoc

def main():
    orders = [1,3]
    for order in orders:
        print("order: ", order)

        eoc, expected_eoc = runTestCurlfree( order )
        checkEOC(eoc, expected_eoc)

main()



# %% [markdown]
# .. index:: Solvers; Saddle Point (Uzawa)
#
# # Saddle Point Solver (using Scipy)
# %%
import matplotlib

matplotlib.rc( 'image', cmap='jet' )
from matplotlib import pyplot
import numpy
from scipy.sparse import linalg
from concurrent.futures import ThreadPoolExecutor

# this will be available in dune.common soon
from dune.grid import cartesianDomain
from dune.alugrid import aluCubeGrid
from ufl import SpatialCoordinate, CellVolume, TrialFunction, TestFunction,\
                inner, dot, div, nabla_grad, grad, dx, as_vector, transpose, Identity, sym, \
                FacetNormal, ds, dS, avg, jump, CellVolume, FacetArea
from dune.ufl import Constant, DirichletBC
import dune.fem
from dune.fem.function import integrate, discreteFunction
from dune.fem.operator import linear as linearOperator

import dune.vem
from dune.vem import vemScheme as vemScheme
from dune.fem.operator import galerkin as galerkinOperator
from dune.fem.scheme import galerkin as galerkinScheme

import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)

def plot(target):
    fig = pyplot.figure(figsize=(10,10))
    target[0].plot(colorbar="vertical", figure=(fig, 211))
    target[1].plot(colorbar="vertical", figure=(fig, 212))
    pyplot.show()

def dgLaplace(beta, p,q, spc, p_bnd, dD):
    n             = FacetNormal(spc)
    he            = avg( CellVolume(spc) ) / FacetArea(spc)
    hbnd          = CellVolume(spc) / FacetArea(spc)
    aInternal     = dot(grad(p), grad(q)) * dx
    diffSkeleton  = beta/he*jump(p)*jump(q)*dS -\
                    dot(avg(grad(p)),n('+'))*jump(q)*dS -\
                    jump(p)*dot(avg(grad(q)),n('+'))*dS
    diffSkeleton -= ( dot(grad(p),n)*q*dD +\
                      p_bnd*dot(grad(q),n)*dD ) * ds
    if p_bnd is not None:
        diffSkeleton += beta/hbnd*(p-p_bnd)*dD*q*ds
    return aInternal + diffSkeleton
class Uzawa:
    def __init__(self, grid, spcU, spcP, dbc_u, forcing,
                 mu, tau, u_h_n=None, explicit=False,
                 tolerance=1e-9, precondition=True, verbose=False,
                 lagrange=False):
        self.dimension = grid.dimension
        self.verbose = verbose
        self.tolerance2 = tolerance**2
        self.dbc_u = dbc_u
        self.mu = Constant(mu, "mu")
        if tau is not None:
            assert tau>0
            self.nu  = Constant(1/tau, "nu")
        else:
            self.nu  = Constant(0, "nu")
        u = TrialFunction(spcU)
        v = TestFunction(spcU)
        p = TrialFunction(spcP)
        q = TestFunction(spcP)
        if u_h_n is not None:
            forcing += self.nu*u_h_n
            if explicit:
                forcing -= dot(u_h_n, nabla_grad(u_h_n))
            else:
                forcing -= ( dot(u_h_n, nabla_grad(u)) + dot(u, nabla_grad(u_h_n)) ) / 2
        epsilon = lambda u: sym(nabla_grad(u))
        mainModel   = ( self.nu*dot(u,v) - dot(forcing,v) + 2*self.mu*inner(epsilon(u), epsilon(v)) ) * dx
        gradModel   = -p*div(v) * dx
        divModel    = -div(u)*q * dx
        massModel   = p*q * dx
        preconModel = inner(grad(p),grad(q)) * dx

        with ThreadPoolExecutor(max_workers=8) as executor:
            if not lagrange:
                mainOpF = executor.submit( vemScheme, ( mainModel==0, *dbc_u ),
                                              gradStabilization=as_vector([self.mu*2,self.mu*2]),
                                              massStabilization=as_vector([self.nu,self.nu])
                                         )
            else:
                mainOpF = executor.submit( galerkinScheme, ( mainModel==0, *dbc_u ) )
            gradOpF = executor.submit(galerkinOperator, gradModel, spcP,spcU)
            divOpF  = executor.submit(galerkinOperator, divModel, spcU,spcP)
            massOpF = executor.submit(galerkinOperator, massModel, spcP)
            if precondition and tau is not None:
                if not lagrange:
                    preconModel = dgLaplace(10, p,q, spcP, 0, 1)
                    preconOpF = executor.submit(galerkinOperator, preconModel, spcP)
                else:
                    preconModel = inner(grad(p),grad(q)) * dx
                    preconOpF = executor.submit(galerkinOperator, (preconModel,DirichletBC(spcP,[0])), spcP)
        self.mainOp, gradOp, divOp, massOp = \
            mainOpF.result(), gradOpF.result(), divOpF.result(), massOpF.result()
        if precondition and tau is not None:
            preconOp = preconOpF.result()
        else:
            preconOp = None

        with ThreadPoolExecutor(max_workers=8) as executor:
            AF = executor.submit(linearOperator,self.mainOp)
            GF = executor.submit(linearOperator,gradOp)
            DF = executor.submit(linearOperator,divOp)
            MF = executor.submit(linearOperator,massOp)
            if preconOp:
                PF = executor.submit(linearOperator,preconOp)
            else:
                PF = None
        self.A, self.G, self.D, self.M = (
          AF.result().as_numpy, GF.result().as_numpy,
          DF.result().as_numpy, MF.result().as_numpy )

        self.Ainv = lambda rhs: linalg.spsolve(self.A,rhs)
        self.Minv = lambda rhs: linalg.spsolve(self.M,rhs)

        if PF:
            self.P = PF.result().as_numpy
            self.Pinv = lambda rhs: linalg.spsolve(self.P,rhs)
        else:
            self.Pinv = None

        self.rhsVelo  = spcU.interpolate(spcU.dimRange*[0], name="vel")
        self.rhsPress = spcP.interpolate(0, name="pres")
        self.rhs_u  = self.rhsVelo.as_numpy
        self.rhs_p  = self.rhsPress.as_numpy
        self.r      = numpy.copy(self.rhs_p)
        self.d      = numpy.copy(self.rhs_p)
        self.precon = numpy.copy(self.rhs_p)
        self.xi     = numpy.copy(self.rhs_u)

    def solve(self,target):
        info = {"uzawa.outer.iterations":0,
                "uzawa.converged":False}
        self.A    = linearOperator(self.mainOp).as_numpy
        self.Ainv = lambda rhs: linalg.spsolve(self.A,rhs)
        velocity = target[0]
        pressure = target[1]
        sol_u = velocity.as_numpy
        sol_p = pressure.as_numpy
        # right hand side for Shur complement problem
        velocity.clear()
        self.mainOp(velocity,self.rhsVelo)
        self.rhs_u *= -1
        self.xi[:]  = self.G*sol_p
        self.rhs_u -= self.xi
        self.mainOp.setConstraints(self.rhsVelo)
        sol_u[:]      = self.Ainv(self.rhs_u[:])
        self.rhs_p[:] = self.D*sol_u
        self.r[:]     = self.Minv(self.rhs_p[:])
        if self.Pinv:
            self.precon.fill(0)
            self.precon[:] = self.Pinv(self.rhs_p[:])
            self.r *= self.mu.value
            self.r += self.nu.value*self.precon
        self.d[:] = self.r[:]
        delta = numpy.dot(self.r,self.rhs_p)
        if self.verbose:
            print(0,delta,self.tolerance2)
        # cg type iteration
        for m in range(1,1000):
            self.xi.fill(0)
            self.rhs_u[:] = self.G*self.d
            self.mainOp.setConstraints([0,]*self.dimension, self.rhsVelo)
            self.xi[:] = self.Ainv(self.rhs_u[:])
            self.rhs_p[:] = self.D*self.xi
            rho = delta / numpy.dot(self.d,self.rhs_p)
            sol_p += rho*self.d
            sol_u -= rho*self.xi
            self.rhs_p[:] = self.D*sol_u
            self.r[:] = self.Minv(self.rhs_p[:])
            if self.Pinv:
                self.precon.fill(0)
                self.precon[:] = self.Pinv(self.rhs_p[:])
                self.r *= self.mu.value
                self.r += self.nu.value*self.precon
            oldDelta = delta
            delta = numpy.dot(self.r,self.rhs_p)
            if self.verbose:
                print(m,delta,self.tolerance2)
            if delta < self.tolerance2:
                info["uzawa.converged"] = True
                break
            gamma = delta/oldDelta
            self.d *= gamma
            self.d += self.r
        info["uzawa.outer.iterations"] = m
        return info

def main(level = 0):
    Lx = 3
    Ly = 1
    N = 2**(level)
    muValue = 1 # 0.01
    tauValue = None # 1e-3
    order = 2
    grid = dune.vem.polyGrid(cartesianDomain([0,0],[Lx,Ly],[N,N]),cubes=False)
    # grid = dune.vem.polyGrid(
    #       dune.vem.voronoiCells([[0,0],[Lx,Ly]], 50*N*N, lloyd=200, load="voronoiseeds")
        #   cartesianDomain([0.,0.],[Lx,Ly],[N,N]), cubes=False
        #   cartesianDomain([0.,0.],[Lx,Ly],[2*N,2*N]), cubes=True
    #    )

    with ThreadPoolExecutor(max_workers=1) as executor:
        spcUF = executor.submit(dune.vem.divFreeSpace, grid, order=order)
        spcPF = executor.submit(dune.vem.bbdgSpace, grid, order=0)
        spcPsmoothF = executor.submit(dune.vem.vemSpace, grid, order=order, testSpaces=[-1,order-1,order-2])
    spcU, spcP, spcPsmooth = spcUF.result(), spcPF.result(), spcPsmoothF.result()
    print(grid.size(0),grid.size(1),grid.size(2)," ",spcU.size,spcP.size)

    x       = SpatialCoordinate(spcU)
    exact_u = as_vector( [x[1] * (1.-x[1]), 0] ) # dy u_x = 1-2y, -dy^2 u_x = 2
    exact_p = (-2*x[0] + 3)*muValue              # dx p   = -2mu
    f       = as_vector( [0,]*grid.dimension )

    # discrete functions
    velocity = spcU.interpolate(spcU.dimRange*[0], name="velocity")
    pressure = spcP.interpolate(0, name="pressure")

    dbc = [ DirichletBC(velocity.space,exact_u,None) ]
    exact_uh = dune.fem.function.uflFunction(grid, order=2, name="exact_uh", ufl=exact_u)
    uzawa = Uzawa(grid, spcU, spcP, dbc, f,
                  muValue, tauValue, u_h_n=exact_uh,
                  tolerance=1e-12, precondition=True, verbose=True)
    print("Setup uzawa",flush=True)
    uzawa.solve([velocity,pressure])

    p,q = TrialFunction(spcPsmooth), TestFunction(spcPsmooth)
    scheme = vemScheme( [ ( inner(grad(p),grad(q)) -
                            div( dot(velocity, nabla_grad(velocity)) - f)*q ) * dx == 0,
                        DirichletBC(spcPsmooth,exact_p) ],
                        gradStabilization=1, massStabilization=None,
                        parameters={"newton.linear.tolerance":1e-14}
                      )
    pressure = spcPsmooth.interpolate(0,name="pRecon")
    scheme.solve(pressure)
    average_p = Constant( pressure.integrate(), "aver_p" )

    edf = as_vector( [inner(grad(v1-v2),grad(v1-v2)) for v1,v2 in zip(velocity,exact_u)] +
                     [((pressure-average_p)-exact_p)**2] )
    err = [ e*e for e in edf ]
    errors  = [ numpy.sqrt(e) for e in integrate(grid, err, order=8) ]

    print(errors, average_p.value)
    # plot([velocity,pressure])
    return errors

if __name__ == "__main__":
    err1 = main(0)
    err2 = main(1)
    assert all( [ e<1e-9 for e in err1 ] )
    assert all( [ e<1e-9 for e in err2 ] )



# %% [markdown]
# .. index:: Solvers; Saddle Point (Uzawa)
#
# # Saddle Point Solver (using Scipy)
# %%
import matplotlib
matplotlib.rc( 'image', cmap='jet' )
from matplotlib import pyplot
import numpy
from scipy.sparse import linalg
from dune.grid import cartesianDomain
from dune.alugrid import aluCubeGrid
from ufl import SpatialCoordinate, CellVolume, TrialFunction, TestFunction,\
                inner, dot, div, nabla_grad, grad, dx, as_vector, transpose, Identity, sym, \
                FacetNormal, ds, dS, avg, jump, CellVolume, FacetArea
from dune.ufl import Constant, DirichletBC
import dune.fem
from dune.fem.operator import linear as linearOperator

import dune.vem
from dune.fem.operator import galerkin as galerkinOperator
from dune.vem import vemScheme as vemScheme
from dune.fem.scheme import galerkin as galerkinScheme

def plot(target):
    fig = pyplot.figure(figsize=(10,10))
    target[0].plot(colorbar="vertical", figure=(fig, 211))
    target[1].plot(colorbar="vertical", figure=(fig, 212))
    pyplot.show()

def dgLaplace(beta, p,q, spc, p_bnd, dD):
    n             = FacetNormal(spc)
    he            = avg( CellVolume(spc) ) / FacetArea(spc)
    hbnd          = CellVolume(spc) / FacetArea(spc)
    aInternal     = dot(grad(p), grad(q)) * dx
    diffSkeleton  = beta/he*jump(p)*jump(q)*dS -\
                    dot(avg(grad(p)),n('+'))*jump(q)*dS -\
                    jump(p)*dot(avg(grad(q)),n('+'))*dS
    diffSkeleton -= ( dot(grad(p),n)*q*dD +\
                      p_bnd*dot(grad(q),n)*dD ) * ds
    if p_bnd is not None:
        diffSkeleton += beta/hbnd*(p-p_bnd)*dD*q*ds
    return aInternal + diffSkeleton
class Uzawa:
    def __init__(self, grid, spcU, spcP, dbc_u, forcing,
                 mu, tau, u_h_n=None, explicit=False,
                 tolerance=1e-9, precondition=True, verbose=False,
                 lagrange=False):
        self.dimension = grid.dimension
        self.verbose = verbose
        self.tolerance2 = tolerance**2
        self.dbc_u = dbc_u
        self.mu = Constant(mu, "mu")
        if tau is not None:
            assert tau>0
            self.nu  = Constant(1/tau, "nu")
        else:
            self.nu  = Constant(0, "nu")
        u = TrialFunction(spcU)
        v = TestFunction(spcU)
        p = TrialFunction(spcP)
        q = TestFunction(spcP)
        forcing = dot(forcing,v)
        if u_h_n is not None:
            forcing += dot(self.nu*u_h_n,v)
            b = lambda u1,u2,phi: dot( dot(u1, nabla_grad(u2)), phi)
            symb = lambda u1,u2,phi: ( b(u1,u2,phi) - b(u1,phi,u2) ) / 2
            if explicit:
                forcing -= b( u_h_n, u_h_n, v )
            else:
                # forcing -= b(u_h_n, u, v)
                # forcing -= b(u,u_h_n,v)
                # ustable forcing -= symb(u_h_n, u, v)
                # default
                forcing -= ( b(u_h_n, u, v) + b(u,u_h_n,v) ) / 2
                # unstable at outflow forcing -= ( symb(u_h_n, u, v) + symb(u,u_h_n,v) ) / 2
        epsilon = lambda u: sym(nabla_grad(u))
        mainModel   = ( self.nu*dot(u,v) - forcing + 2*self.mu*inner(epsilon(u), epsilon(v)) ) * dx
        gradModel   = -p*div(v) * dx
        divModel    = -div(u)*q * dx
        massModel   = p*q * dx
        preconModel = inner(grad(p),grad(q)) * dx

        if not lagrange:
            self.mainOp = vemScheme( ( mainModel==0, *dbc_u ),
                                       gradStabilization=as_vector([self.mu*2,self.mu*2]),
                                       massStabilization=as_vector([self.nu,self.nu])
                                     )
        else:
            self.mainOp = galerkinScheme( ( mainModel==0, *dbc_u ) )
        gradOp    = galerkinOperator( gradModel, spcP,spcU)
        divOp     = galerkinOperator( divModel, spcU,spcP)
        massOp    = galerkinOperator( massModel, spcP)

        self.mainLinOp = linearOperator(self.mainOp)
        self.G    = linearOperator(gradOp).as_numpy
        self.D    = linearOperator(divOp).as_numpy
        self.M    = linearOperator(massOp).as_numpy
        self.Minv = lambda rhs: linalg.spsolve(self.M,rhs)

        if precondition and self.mainOp.model.nu > 0:
            if not lagrange:
                preconModel = dgLaplace(10, p,q, spcP, 0, 1)
                preconOp    = galerkinOperator( preconModel, spcP)
            else:
                preconModel = inner(grad(p),grad(q)) * dx
                preconOp    = galerkinOperator( (preconModel,DirichletBC(spcP,0)), spcP)
            self.P    = linearOperator(preconOp).as_numpy
            self.Pinv = lambda rhs: linalg.spsolve(self.P,rhs)
        else:
            self.Pinv = None

        self.rhsVelo  = spcU.interpolate(spcU.dimRange*[0], name="vel")
        self.rhsPress = spcP.interpolate(0, name="pres")
        self.rhs_u  = self.rhsVelo.as_numpy
        self.rhs_p  = self.rhsPress.as_numpy
        self.r      = numpy.copy(self.rhs_p)
        self.d      = numpy.copy(self.rhs_p)
        self.precon = numpy.copy(self.rhs_p)
        self.xi     = numpy.copy(self.rhs_u)

    def solve(self,target):
        velocity = target[0]
        pressure = target[1]
        info = {"uzawa.outer.iterations":0,
                "uzawa.converged":False}
        # problem is linear but coefficients depend on previous time step so need to reassemble
        self.mainOp.jacobian(velocity, self.mainLinOp)
        A = self.mainLinOp.as_numpy
        Ainv = lambda rhs: linalg.spsolve(A,rhs)
        sol_u = velocity.as_numpy
        sol_p = pressure.as_numpy
        # right hand side for Shur complement problem
        velocity.clear()
        self.mainOp(velocity,self.rhsVelo)
        self.rhs_u *= -1
        self.xi[:]  = self.G*sol_p
        self.rhs_u -= self.xi
        self.mainOp.setConstraints(self.rhsVelo)
        sol_u[:]      = Ainv(self.rhs_u[:])
        self.rhs_p[:] = self.D*sol_u
        self.r[:]     = self.Minv(self.rhs_p[:])
        if self.Pinv:
            self.precon.fill(0)
            self.precon[:] = self.Pinv(self.rhs_p[:])
            self.r *= self.mu.value
            self.r += self.nu.value*self.precon
        self.d[:] = self.r[:]
        delta = numpy.dot(self.r,self.rhs_p)
        if self.verbose:
            print(0,delta,self.tolerance2)
        # cg type iteration
        for m in range(1,1000):
            self.xi.fill(0)
            self.rhs_u[:] = self.G*self.d
            self.mainOp.setConstraints([0,]*self.dimension, self.rhsVelo)
            self.xi[:] = Ainv(self.rhs_u[:])
            self.rhs_p[:] = self.D*self.xi
            rho = delta / numpy.dot(self.d,self.rhs_p)
            sol_p += rho*self.d
            sol_u -= rho*self.xi
            self.rhs_p[:] = self.D*sol_u
            self.r[:] = self.Minv(self.rhs_p[:])
            if self.Pinv:
                self.precon.fill(0)
                self.precon[:] = self.Pinv(self.rhs_p[:])
                self.r *= self.mu.value
                self.r += self.nu.value*self.precon
            oldDelta = delta
            delta = numpy.dot(self.r,self.rhs_p)
            if self.verbose:
                print(m,delta,self.tolerance2)
            if delta < self.tolerance2:
                info["uzawa.converged"] = True
                break
            gamma = delta/oldDelta
            self.d *= gamma
            self.d += self.r
        info["uzawa.outer.iterations"] = m
        return info

def main():
    useVem   = True
    muValue  = 0.1
    tauValue = 1e-2
    order    = 2
    grid = dune.vem.polyGrid(cartesianDomain([0,0],[3,1],[30*4,10*4]),cubes=False)
    if useVem:
        spcU = dune.vem.divFreeSpace( grid, order=order, conforming=True)
        spcP = dune.vem.bbdgSpace( grid, order=0 )
    else:
        spcU = dune.fem.space.lagrange( grid, order=order, dimRange=2 )
        spcP = dune.fem.space.lagrange( grid, order=order-1 )

    x       = SpatialCoordinate(spcU)
    exact_u = as_vector( [x[1] * (1.-x[1]), 0] ) # dy u_x = 1-2y, -dy^2 u_x = 2
    exact_p = (-2*x[0] + 2)*muValue              # dx p   = -2mu
    f       = as_vector( [0,]*grid.dimension )
    f      += exact_u/tauValue

    # discrete functions
    velocity = spcU.interpolate(spcU.dimRange*[0], name="velocity")
    pressure = spcP.interpolate(0, name="pressure")

    dbc = [ DirichletBC(velocity.space,exact_u,None) ]
    uzawa = Uzawa(grid, spcU, spcP, dbc, f,
                  muValue, tauValue, u_h_n=None,
                  tolerance=1e-9, precondition=True, verbose=True, # tolerance 2e-5 (TH p=2)
                  lagrange=not useVem )
    uzawa.solve([velocity,pressure])
    print("Retry")
    uzawa.solve([velocity,pressure])

    fig = pyplot.figure(figsize=(10,10))
    velocity.plot(colorbar="vertical", figure=(fig, 211))
    pressure.plot(colorbar="vertical", figure=(fig, 212))
    pyplot.show()
    fig = pyplot.figure(figsize=(10,10))
    dune.fem.plotting.plotPointData(grad(velocity[0])[0], grid=grid,colorbar="vertical", figure=(fig, 211))
    dune.fem.plotting.plotPointData(grad(velocity[0])[1], grid=grid,colorbar="vertical", figure=(fig, 212))
    pyplot.show()

if __name__ == "__main__":
    main()



import pickle
from matplotlib import pyplot
import numpy
results1 = pickle.load(open('conditioning1A.dump','rb'))
results2 = pickle.load(open('conditioning2A.dump','rb'))
results3 = pickle.load(open('conditioning3.dump','rb'))
resultsL = pickle.load(open('conditioningLagrange.dump','rb'))

x = []
N = []
for i,r in enumerate(results1):
    N.append(r[0])
    x.append(r[1])
x = list(set(x))
N = list(set(N))
x.sort()
N.sort()
y1 = numpy.zeros((len(N),len(x)))
y2 = numpy.zeros((len(N),len(x)))
y3 = numpy.zeros((len(N),len(x)))
yL = numpy.zeros((len(N),len(x)))

for r1,r2,r3,rL in zip(results1,results2,results3,resultsL):
    n = N.index(r1[0])
    o = x.index(r1[1])
    # y1[n][o] = r1[5][1]
    y1[n][o] = r1[6]["conditioning"]
    # y2[n][o] = r2[5][1]
    y2[n][o] = r2[6]["conditioning"]
    # y3[n][o] = r3[5][1]
    y3[n][o] = r3[6]["conditioning"]
    # yL[n][o] = rL[5][1]
    yL[n][o] = rL[6]["conditioning"]

for i in range(len(y1)):
    # pyplot.semilogy(x,y1[i])
    pyplot.semilogy(x,y2[i])
    pyplot.semilogy(x,y3[i])
    pyplot.semilogy(x,yL[i])
pyplot.show()



import dune.vem
import matplotlib
matplotlib.rc( 'image', cmap='jet' )
from matplotlib import pyplot
import math, numpy
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem.function import integrate, discreteFunction
import dune.fem

from ufl import *
import dune.ufl

dune.fem.parameter.append({"fem.verboserank": 0})
parameters = {"newton.linear.tolerance": 1e-8,
              "newton.tolerance": 5e-6,
              "newton.lineSearch": "simple",
              "newton.linear.verbose": False,
              "newton.verbose": True
              }
order, Lx,Ly = 3,  3,1
# order, Lx,Ly = 5,  1,1.1
mass = dune.ufl.Constant(1, "mu")
D = dune.ufl.Constant(1, "D")

x = SpatialCoordinate(triangle)
def model(space):
    u = TrialFunction(space)
    v = TestFunction(space)

    if False:
        # ( -y+x^2y , x-y^2x)
        # div: 2xy - 2yx = 0
        exact = as_vector([-x[1]+x[0]**2*x[1],
                            x[0]-x[1]**2*x[0]])
    elif True:
        # div: (p+1)x^py^p - (p+1)y^px^p = 0
        p = 1
        exact = as_vector([-x[1]+x[0]**(p+1)*x[1]**p,
                            x[0]-x[1]**(p+1)*x[0]**p])
    else:
        exact = as_vector( [x[1] * (1.-x[1]), 0] )
        # exact = as_vector( [10.,0.] )

    a = (D*inner(grad(u),grad(v)) + mass*dot(u,v) ) * dx

    b = dot( -D*div(grad(exact)) + mass*exact, v) * dx
    dbc = [dune.ufl.DirichletBC(space, exact, i+1) for i in range(4)]
    return a,b,dbc,exact

oldErrors = []
oldDiams = None
for i in range(0,6):
    errors = []
    N = 2**i # 2**(i+1)
    polyGrid = dune.vem.polyGrid(
          dune.vem.voronoiCells([[0,0],[Lx,Ly]], 10*N*N, lloyd=200, fileName="test", load=True)
          # cartesianDomain([0.,0.],[3,1],[3*N,2*N]), cubes=False
          # cartesianDomain([0.,0.],[3,1],[3*N,3*N]), cubes=True
      )

    indexSet = polyGrid.indexSet
    @gridFunction(polyGrid, name="cells")
    def polygons(en,x):
        return polyGrid.hierarchicalGrid.agglomerate(indexSet.index(en))
    # polygons.plot()
    space = dune.vem.divFreeSpace( polyGrid, order=order)
    a,b,dbc,exact = model(space)
    dfI = space.interpolate(exact,name="interpol")
    if False:
        df = space.interpolate(exact,name="solution")
        df.plot()
    else:
        df = space.interpolate(exact,name="solution")
        scheme = dune.vem.vemScheme(
                  [a==b, *dbc], space, solver=("suitesparse","umfpack"),
                  parameters=parameters,
                  gradStabilization=D,
                  massStabilization=mass)
        info = scheme.solve(target=df)
        df.plot()

    edf = exact-df
    err = [inner(edf,edf), inner(grad(edf),grad(edf))]
    errors += [ numpy.sqrt(e) for e in integrate(polyGrid, err, order=8) ]

    print(errors,"#",space.diameters(),flush=True)
    if len(oldErrors)>0:
        factor = oldDiams[0] / space.diameters()[0]
        print(i,"eocs:", [ math.log(oe/e)/math.log(factor)
                for oe,e in zip(oldErrors,errors) ],flush=True)
    oldErrors = errors
    oldDiams = space.diameters()
    vtk = space.gridView.writeVTK("cos"+"_"+str(i),subsampling=3,
      celldata={"solution":df,"error":edf,
                "cells":polygons,
               })



#!/usr/bin/python

# Compute the convex hull of a set of 2D points
# A Python implementation of the qhull algorithm
#
# Tested with Python 2.6.5 on Ubuntu 10.04.4

# Copyright (c) 2008 Dave (www.literateprograms.org)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import division
import numpy as np

link = lambda a,b: np.concatenate((a,b[1:]))
edge = lambda a,b: np.concatenate(([a],[b]))

def qhull2D(sample):
    def dome(sample,base):
        h, t = base
        dists = np.dot(sample-h, np.dot(((0,-1),(1,0)),(t-h)))
        outer = np.repeat(sample, dists>1e-16, 0)
        if len(outer):
            pivot = sample[np.argmax(dists)]
            # print("dome(pivot):",pivot)
            return link(dome(outer, edge(h, pivot)),
                        dome(outer, edge(pivot, t)))
        else:
            return base
    if len(sample) > 2:
        axis = sample[:,0]
        base = np.take(sample, [np.argmin(axis), np.argmax(axis)], 0)
        return link(dome(sample, base), dome(sample, base[::-1]))
    else:
        return sample



# %% [markdown]
# .. index:: Methods; Virtual Finite Elements
#
# The main concepts of our virtual element implementation is provided in
# <cite data-cite="VEM"></cite>
# where we focus on the fourth order problems but the second order problems
# are discussed there as well.

# # Elliptic Problems
#
# We first consider an elliptic problem with varying coefficients and Dirichlet boundary conditions
# \begin{align*}
# -\nabla D(x)\nabla u + \mu(x) u &= f, && \text{in } \Omega, \\
# u &= g, && \text{on } \partial\Omega,
# \end{align*}
# with $\Omega=[-\frac{1}{2},1]^2$ and choosing the forcing and the boundary conditions
# so that the exact solution is equal to
# \begin{align*}
# u(x,y) &= xy\cos(\pi xy)
# \end{align*}

# First some setup code:
# %%
try:
    import dune.vem
except:
    print("This example needs 'dune.vem'")
    import sys
    sys.exit(0)
import matplotlib
matplotlib.rc( 'image', cmap='jet' )
from matplotlib import pyplot
import numpy
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem.function import integrate, discreteFunction
import dune.fem

from ufl import *
import dune.ufl

dune.fem.parameter.append({"fem.verboserank": 0})

# %% [markdown]
# We use a grid build up of voronoi cells around $50$ random points
# in the interval $[-\frac{1}{2},1]\times [-\frac{1}{2},1]$ using 100
# iterations of Lloyd's algorithm to improve the quality of the grid.

# %%
polyGrid = dune.vem.polyGrid( dune.vem.voronoiCells([[-0.5,-0.5],[1,1]], 50, lloyd=100) )

# %% [markdown]
# One can also use a standard simplex or cube grid, e.g.,
# polyGrid = dune.vem.polyGrid( cartesianDomain([-0.5,-0.5],[1,1],[10,10]), cubes=False)
#
# In general we can construct a `polygrid` by providing a dictionary with
# the `vertices` and the `polygons`. The `voronoiCells` function creates
# such a dictionary using random seeds to generate voronoi cells which are
# cut off using the provided `cartesianDomain`. The seeds can be
# provided as list of points as second argument:
# ```
# voronoiCells(constructor, towers, fileName=None, load=False):
# ```
# If a `fileName` is provided the seeds will be written to disc or if that
# file exists they will be loaded from that file if `load=True`,
# to make results reproducible.
#
# As an example an output of `voronoiCells(constructor,5)` is
# ```
# {'polygons': [ [4, 5, 2, 3], [ 8, 10,  9,  7], [7, 9, 1, 3, 4],
#                [11, 10,  8,  0], [8, 0, 6, 5, 4, 7] ],
#  'vertices': [ [ 0.438, 1.  ],    [ 1. , -0.5 ],
#                [-0.5, -0.5  ],    [ 0.923, -0.5 ],
#                [ 0.248,  0.2214], [-0.5,  0.3027],
#                [-0.5,  1. ],      [ 0.407,0.4896],
#                [ 0.414,  0.525],  [ 1.,  0.57228],
#                [ 1., 0.88293],    [ 1.,  1. ] ] }
# ```
#
# Let's take a look at the grid with the 50 polygons triangulated

# %%
indexSet = polyGrid.indexSet
@gridFunction(polyGrid, name="cells")
def polygons(en,x):
    return polyGrid.hierarchicalGrid.agglomerate(indexSet.index(en))
# polygons.plot(colorbar="horizontal")

# %% [markdown]
# The vem space is now setup in exactly the same way as usual but the type
# of space constructed is defined by the final argument which defines the
# moments used on the subentities of a given codimension. So
# `testSpaces=[-1,order-1,order-2]` means: use no vertex values (-1),
# order-1 moments on the edges and order-2 moments in the inside. So this
# gives us a non-conforming space for second order problems - while using
# `testSpaces=[0,order-2,order-2]` defines a conforming space.

# %%
order = 3
space = dune.vem.vemSpace( polyGrid, order=order, dimRange=1, storage="numpy",
                           testSpaces=[-1,order-1,order-2])

# %% [markdown]
# Now we define the model starting with the exact solution:

# %%
x = SpatialCoordinate(space)
u = TrialFunction(space)
v = TestFunction(space)

exact = as_vector( [x[0]*x[1] * cos(pi*x[0]*x[1])] )

massCoeff = 1+sin(dot(x,x))       # factor for mass term
diffCoeff = 11-0.9*cos(dot(x,x))   # factor for diffusion term

a = (diffCoeff*inner(grad(u),grad(v)) + massCoeff*dot(u,v) ) * dx

# finally the right hand side and the boundary conditions
b = (-div(diffCoeff*grad(exact[0])) + massCoeff*exact[0] ) * v[0] * dx
dbc = [dune.ufl.DirichletBC(space, exact, i+1) for i in range(4)]

# %% [markdown]
# Finally we can construct the solver passing in the space the pde
# description and arguments for stabilization:

# %%
parameters = {"newton.linear.tolerance": 1e-12,
              "newton.linear.preconditioning.method": "jacobi",
              "penalty": 10*order*order,  # for the dg schemes
              "newton.linear.verbose": False,
              "newton.verbose": False
              }
df = space.interpolate([0],name="solution")
scheme = dune.vem.vemScheme( [a==b, *dbc], space, solver="cg",
                             gradStabilization=diffCoeff,
                             massStabilization=massCoeff,
                             parameters=parameters )
info = scheme.solve(target=df)
# df.interpolate(exact)
print("size of space:",space.size,flush=True)
df.plot(level=3)
plot(df-exact, grid=polyGrid, gridLines=None, level=3)
edf = exact-df
err = [inner(edf,edf),
       inner(grad(edf),grad(edf))]
errors = [ numpy.sqrt(e) for e in integrate(polyGrid, err, order=8) ]
print(errors)

# %% [markdown]
# Repeating the same test with a H^1-conforming space

# %%
space = dune.vem.vemSpace( polyGrid, order=order, dimRange=1, storage="istl",
                           testSpaces=[0,order-2,order-2])
df = space.interpolate([0],name="solution")
scheme = dune.vem.vemScheme( [a==b, *dbc], space, solver="cg",
                             gradStabilization=diffCoeff,
                             massStabilization=massCoeff,
                             parameters=parameters )
info = scheme.solve(target=df)
# df.interpolate(exact)
print("size of space:",space.size,flush=True)
df.plot(level=3)
plot(df-exact, grid=polyGrid, gridLines=None, level=3)
edf = exact-df
err = [inner(edf,edf),
       inner(grad(edf),grad(edf))]
errors = [ numpy.sqrt(e) for e in integrate(polyGrid, err, order=8) ]
print(errors)

# %% [markdown]
# we can compare different method, e.g., a lagrange/dg scheme (on the the subtriangulation),
# a bounding box dg method and conforming/non conforming VEM:

# %%
methods = [ ### "[legend,space,scheme,spaceKwargs,schemeKwargs]"
            ["lagrange",
             dune.fem.space.lagrange,dune.fem.scheme.galerkin,{},{}],
            ["dg",
             dune.fem.space.dgonb, dune.fem.scheme.dg,  {}, {"penalty":diffCoeff}],
            ["vem-conforming",
             dune.vem.vemSpace,    dune.vem.vemScheme,
                {"testSpaces":[0,order-2,order-2]},  # conforming vem space
                {"gradStabilization":diffCoeff, "massStabilization":massCoeff}],
            ["vem-nonconforming",
             dune.vem.vemSpace,    dune.vem.vemScheme,
                 {"testSpaces":[-1,order-1,order-2]},  # non-conforming vem space
                 {"gradStabilization":diffCoeff, "massStabilization":massCoeff}],
            ["bb-dg",
             dune.vem.bbdgSpace,   dune.vem.bbdgScheme, {}, {"penalty":diffCoeff}],
   ]

# %% [markdown]
# We now define a function to compute the solution and the $L^2,H^1$ error
# given a grid and a space

# %%
def compute(grid, space,spaceArgs, schemeName,schemeArgs):
    space = space( grid, order=order, dimRange=1, storage="istl", **spaceArgs )
    df = space.interpolate([0],name="solution")
    scheme = schemeName( [a==b, *dbc], space, solver="cg", **schemeArgs,
                         parameters=parameters )
    info = scheme.solve(target=df)

    # compute the error
    edf = exact-df
    err = [inner(edf,edf),
           inner(grad(edf),grad(edf))]
    errors = [ numpy.sqrt(e) for e in integrate(grid, err, order=8) ]

    return df, errors, info

# %% [markdown]
# Finally we iterate over the requested methods and solve the problems

# %%
fig = pyplot.figure(figsize=(5*len(methods),10))
figPos = 111+10*len(methods)
for i,m in enumerate(methods):
    dfs,errors,info = compute(polyGrid, m[1],m[3], m[2],m[4])
    print("method (",m[0],"):",
          "Size: ",dfs.space.size, "L^2: ", errors[0], "H^1: ", errors[1],
          info["linear_iterations"], flush=True)
    dfs.plot(figure=(fig,figPos+i), gridLines=None, colorbar="horizontal")

# %% [markdown]
# # Nonlinear Elliptic Problem
# We can easily set up a non linear problem

# %%
space = dune.vem.vemSpace( polyGrid, order=1, dimRange=1, storage="istl", conforming=True )
u = TrialFunction(space)
v = TestFunction(space)
x = SpatialCoordinate(space)
exact = as_vector ( [  (x[0] - x[0]*x[0] ) * (x[1] - x[1]*x[1] ) ] )
Dcoeff = lambda u: 1.0 + u[0]**2
a = (Dcoeff(u) * inner(grad(u), grad(v)) ) * dx
b = -div( Dcoeff(exact) * grad(exact[0]) ) * v[0] * dx
dbcs = [dune.ufl.DirichletBC(space, exact, i+1) for i in range(4)]
scheme = dune.vem.vemScheme( [a==b, *dbcs], space, gradStabilization=Dcoeff(u),
                             solver="cg", parameters=parameters)
solution = space.interpolate([0], name="solution")
info = scheme.solve(target=solution)
edf = exact-solution
errors = [ numpy.sqrt(e) for e in
           integrate(polyGrid, [inner(edf,edf),inner(grad(edf),grad(edf))], order=5) ]
print("non linear problem:", errors )
solution.plot(gridLines=None)

# %% [markdown]
# # Linear Elasticity
# Next we solve a linear elasticity equation using a conforming VEM space:

# First we setup the domain
# %%
L, W = 1, 0.2

beamGrid = dune.vem.polyGrid( dune.vem.voronoiCells([[0,0],[L,W]], 120) )
indexSet = beamGrid.indexSet
@gridFunction(beamGrid, name="cells")
def polygons(en,x):
    return beamGrid.hierarchicalGrid.agglomerate(indexSet.index(en))
polygons.plot(colorbar="horizontal")

# instead of providing the moments we can simply add a parameter 'conforming' to construct the H^1-conforming space
space = dune.vem.vemSpace( beamGrid, order=2, dimRange=2, storage="istl", conforming=True)

# %% [markdown]

# %%
# some model constants
mu = 1
rho = 1
delta = W/L
gamma = 0.4*delta**2
beta = 1.25
lambda_ = beta
g = gamma

# clamped boundary on the left
x = SpatialCoordinate(space)
dbc = dune.ufl.DirichletBC(space, as_vector([0,0]), x[0]<1e-10)

# Define strain and stress
def epsilon(u):
    return 0.5*(nabla_grad(u) + nabla_grad(u).T)
def sigma(u):
    return lambda_*nabla_div(u)*Identity(2) + 2*mu*epsilon(u)

# Define the variational problem
u = TrialFunction(space)
v = TestFunction(space)
f = dune.ufl.Constant((0, -rho*g))
a = inner(sigma(u), epsilon(v))*dx
b = dot(as_vector([0,-rho*g]),v)*dx

# Compute solution
displacement = space.interpolate([0,0], name="displacement")
scheme = dune.vem.vemScheme( [a==b, dbc], space,
        gradStabilization = as_vector([lambda_+2*mu, lambda_+2*mu]),
        solver="cg", parameters=parameters )
info = scheme.solve(target=displacement)

# %% [markdown]
# Show the magnitude of the displacement field, stress and the deformed beam

# %%
fig = pyplot.figure(figsize=(20,10))
displacement.plot(gridLines=None, figure=(fig, 121), colorbar="horizontal")
s = sigma(displacement) - (1./3)*tr(sigma(displacement))*Identity(2)
von_Mises = sqrt(3./2*inner(s, s))
plot(von_Mises, grid=beamGrid, gridLines=None, figure=(fig, 122), colorbar="horizontal")

# %% [markdown]
# Finally, we plot the deformed beam

# %%
from dune.fem.view import geometryGridView
position = space.interpolate( x+displacement, name="position" )
beam = geometryGridView( position )
beam.plot()

# %% [markdown]
# # Fourth order problem
# As final example we solve some fourth order PDEs usign a non-conforming VEM
# space for $H^2$ functions. To construct the space we just need to define
# a suitable 'moments' vector to construct a suitable space for $H^2$ problems.
#
# __Note__: at the moment a $H^2$-conforming space is not implemented.

# %%
ncC1testSpaces = [ [0], [order-3,order-2], [order-4] ]

# %% [markdown]
# We test the method using a biharmonic problem.
# \begin{align*}
# -\Delta^2 u &= f, && \text{in } \Omega, \\
# u &= g, && \text{on } \partial\Omega, \\
# \nabla u.n &= 0, && \text{on } \partial\Omega,
# \end{align*}
#
# __Note__:
#       For functions with continuous derivatives we have
#       laplace(u)*laplace(v)*dx = inner(u,v)*dx
#       as can be seen by using integration by parts on the mixed terms on the right
#       and using continuity of u,v.
#       For the non-conforming spaces we don't have continuity of the
#       derivatives so the equivalence does not hold and one should use the
#       right hand side directly to obtain a coercive bilinear form w.r.t.
#       the norm on $H^2$ (the left is not a norm in this case).
#       For computing the forcing term 'b' both formula are fine since
#       'exact' is smooth enough.

# %%
polyGrid = dune.vem.polyGrid( dune.vem.voronoiCells([[0,0],[1,1]], 50, lloyd=100) )
space = dune.vem.vemSpace( polyGrid, order=order, dimRange=1, storage="istl",
                           testSpaces=ncC1testSpaces)

x = SpatialCoordinate(space)
exact = as_vector( [sin(2*pi*x[0])**2*sin(2*pi*x[1])**2] )

laplace = lambda w: div(grad(w))
H = lambda w: grad(grad(w))
u = TrialFunction(space)
v = TestFunction(space)
a = ( inner(H(u[0]),H(v[0])) ) * dx

# finally the right hand side and the boundary conditions
b = laplace(laplace(exact[0])) * v[0] * dx
dbcs = [dune.ufl.DirichletBC(space, [0], i+1) for i in range(4)]

scheme = dune.vem.vemScheme( [a==b, *dbcs], space, hessStabilization=1,
                             solver="cg", parameters=parameters )

# solution = space.interpolate([0], name="solution") # issue here for C^1 spaces
solution = discreteFunction(space, name="solution")
info = scheme.solve(target=solution)
edf = exact-solution
errors = [ numpy.sqrt(e) for e in
           integrate(polyGrid, [inner(edf,edf),inner(grad(edf),grad(edf))], order=5) ]
print("bi-laplace errors:", errors )
solution.plot(gridLines=None)



from dune import create
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem.function import integrate, discreteFunction
from dune.fem import parameter
from dune.fem.operator import linear as linearOperator

import ufl.algorithms
from ufl import *
import dune.ufl

def interpolate_secondorder(space, exact):
    df = space.interpolate(exact,name="solution")

    edf = exact-df
    err = [inner(edf,edf),
           inner(grad(edf),grad(edf))]

    return err

def interpolate_fourthorder(space, exact):
    df = space.interpolate(exact,name="solution")

    edf = exact-df
    err = [inner(edf,edf),
            inner(grad(edf),grad(edf)),
            inner(grad(grad(edf)),grad(grad(edf)))]

    return err


import matplotlib
matplotlib.rc( 'image', cmap='jet' )
from matplotlib import pyplot
import math
from dune import create
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem.function import integrate, discreteFunction
from dune.fem import parameter
from dune.vem import voronoiCells, divFreeSpace, vemScheme
from dune.fem.operator import linear as linearOperator
from scipy.sparse.linalg import spsolve
from pprint import pprint

from ufl import *
import dune.ufl

maxLevel     = 4
order        = 2
level        = 5
constructor = cartesianDomain([0,0],[1,1],[2**level,2**level])
polyGrid = create.grid("agglomerate", constructor, cubes=False )
space = divFreeSpace(polyGrid, order=order, storage="numpy")
x = SpatialCoordinate(space)
uh = space.interpolate([x[1],-x[0]],name="hallo")
uh.plot()



import matplotlib
matplotlib.rc( 'image', cmap='jet' )
from matplotlib import pyplot
import math
from dune import create
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem.function import integrate, discreteFunction
from dune.fem import parameter
from dune.vem import voronoiCells
from dune.fem.operator import linear as linearOperator
from scipy.sparse.linalg import spsolve
import numpy

import ufl.algorithms
from ufl import *
import dune.ufl

dune.fem.parameter.append({"fem.verboserank": -1})

maxLevel = 5
order = 3

parameters = {"newton.linear.tolerance": 1e-12,
              "newton.linear.preconditioning.method": "jacobi",
              "penalty": 40,  # for the dg schemes
              "newton.linear.verbose": False,
              "newton.verbose": False
              }

def runTest(exact, spaceConstructor, get_df):
    Nval = 5
    N_values = [2**i*Nval for i in range(maxLevel)]
    results = []
    for N in N_values:
        print('grid size N:', N)

        # set up grid for testing
        constructor = cartesianDomain([0,0],[1,1],[N,N])
        cells = voronoiCells(constructor,N*N,"voronoiseeds",load=True,show=False,lloyd=10)
        grid = create.grid("agglomerate", cells, convex=True)

        # get dimension of range
        dimRange = exact.ufl_shape[0]
        print('dim Range:', dimRange)

        # construct space to use using spaceConstructor passed in
        space = spaceConstructor(grid,dimRange)

        # get sizes for eoc calculation
        sizes = [grid.hierarchicalGrid.agglomerate.size, space.size, *space.diameters()]

        # solve for df
        df = get_df(space,exact)

        # calculate errors
        edf = exact-df
        err = [inner(edf,edf),
           inner(grad(edf),grad(edf)),
           inner(grad(grad(edf)),grad(grad(edf))),
           energy]
        errors = [ math.sqrt(e) for e in integrate(grid, err, order=8) ]

        print("bi-laplace errors:", errors )
        results += [[sizes] + [errors]]

    # calculate eocs
    h = lambda sizes: sizes[3]
    eoc = len(results[0][1])*[-1]
    for level in range(0,maxLevel):
        print(*results[level], *eoc)
        if level < maxLevel -1:
            eoc = [math.log(results[level+1][1][j]/results[level][1][j])/
            math.log(h(results[level+1][0])/h(results[level][0]))
            for j in range(len(eoc))]

    assert(0.8*(order-1) <= eoc[2] <= 1.2*(order-1), "eoc out of expected range")

def testElliptic(space, exact):
    # set up scheme, df, and solve
    u = TrialFunction(space)
    v = TestFunction(space)
    x = SpatialCoordinate(space)

    massCoeff = 1+sin(dot(x,x))       # factor for mass term
    diffCoeff = 1-0.9*cos(dot(x,x))

    a = (diffCoeff*inner(grad(u),grad(v)) + massCoeff*dot(u,v) ) * dx

    # finally the right hand side and the boundary conditions
    b = (-div(diffCoeff*grad(exact[0])) + massCoeff*exact[0] ) * v[0] * dx
    dbc = [dune.ufl.DirichletBC(space, exact, i+1) for i in range(4)]

    df = space.interpolate([0],name="solution")
    scheme = dune.vem.vemScheme( [a==b, *dbc], space, solver="cg",
                            gradStabilization=diffCoeff,
                            massStabilization=massCoeff,
                            parameters=parameters )

    scheme.solve(target=df)

    return df

def biharmonic(space, exact):
    laplace = lambda w: div(grad(w))
    H = lambda w: grad(grad(w))
    u = TrialFunction(space)
    v = TestFunction(space)

    a = ( inner(H(u[0]),H(v[0])) ) * dx
    b = ( laplace(laplace(exact[0])) )* v[0] * dx
    dbcs = [dune.ufl.DirichletBC(space, [0], i+1) for i in range(4)]

    scheme = dune.vem.vemScheme( [a==b, *dbcs], space, hessStabilization=1, solver="cg", parameters=parameters )
    df = discreteFunction(space, name="solution")
    info = scheme.solve(target=df)

    return df

def main():
    # test with conforming second order VEM space
    conformingSpaceConstructor = lambda grid, r: dune.vem.vemSpace( grid, order=order, dimRange=r, storage="istl", testSpaces = [0,order-2,order-2] )
    # define exact solution
    x = ufl.SpatialCoordinate(ufl.triangle)
    exact = as_vector( [x[0]*x[1] * cos(pi*x[0]*x[1])] )

    # run tests for second order case
    # runTest(exact, conformingSpaceConstructor, testElliptic)

    # exact solution for biharmonic problem
    exact_biharmonic = as_vector( [sin(2*pi*x[0])**2*sin(2*pi*x[1])**2] )
    ncC1testSpaces = [ [0], [order-3,order-2], [order-4] ]
    nonconformingSpaceConstructor = lambda grid, r: dune.vem.vemSpace( grid, order=order, dimRange=r, storage="istl", testSpaces = ncC1testSpaces )

    # run tests for fourth order problems
    runTest(exact_biharmonic, nonconformingSpaceConstructor, biharmonic)

main()


from dune import create
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem.function import integrate, discreteFunction
from dune.fem import parameter
from dune.vem import voronoiCells
from dune.fem.operator import linear as linearOperator

import ufl.algorithms
from ufl import *
import dune.ufl

from script import runTest, checkEOC, interpolate
from interpolate import interpolate_fourthorder

def perturbation(space, exact):
    epsilon = 1e-8

    laplace = lambda w: div(grad(w))
    H = lambda w: grad(grad(w))
    u = TrialFunction(space)
    v = TestFunction(space)

    kappa = epsilon*epsilon
    beta  = 1
    gamma = 0
    laplaceCoeff  = 1
    mu            = 1

    # bilinear form
    a = ( kappa*inner(H(u[0]),H(v[0])) +\
      laplaceCoeff*beta*inner(grad(u),grad(v)) +\
      mu*gamma*inner(u,v)
    ) * dx

    # right hand side and the boundary conditions
    q = sum([ H(kappa*H(exact[0])[i,j])[i,j] for i in range(2) for j in range(2) ])
    b = ( q -\
        laplaceCoeff*div(beta*grad(exact[0])) +\
        mu*gamma*exact[0] ) * v[0] * dx
    dbc = [dune.ufl.DirichletBC(space, [0], i+1) for i in range(4)]

    # stability coefficients
    biLaplaceCoeff = kappa
    diffCoeff      = laplaceCoeff*beta
    massCoeff      = mu*gamma

    df = discreteFunction(space, name="solution")
    info = {"linear_iterations":1}

    scheme = dune.vem.vemScheme( [a==b, *dbc], space,
                            hessStabilization=biLaplaceCoeff,
                            gradStabilization=diffCoeff,
                            massStabilization=massCoeff,
                          )

    scheme.solve(target=df)

    edf = exact-df
    energy  = replace(a,{u:edf,v:edf}).integrals()[0].integrand()
    err = [energy]

    return err

def runTestPerturbation(testSpaces, order):
    x = SpatialCoordinate(triangle)
    exact = as_vector( [sin(2*pi*x[0])**2*sin(2*pi*x[1])**2] )
    spaceConstructor = lambda grid, r: dune.vem.vemSpace( grid,
                                                          order=order,
                                                          dimRange=r,
                                                          testSpaces=testSpaces )

    if (interpolate()):
      eoc = runTest(exact, spaceConstructor, interpolate_fourthorder)
      expected_eoc = [order+1, order, order-1]
    else:
      eoc = runTest(exact, spaceConstructor, perturbation)
      expected_eoc = [order]

    return eoc, expected_eoc

def main():
  orders = [2,4]
  for order in orders:
    print("order: ", order)
    C1C0testSpaces = [ [0], [order-2, order-2], [order-4] ]
    print("C1 C0 test spaces: ", C1C0testSpaces)

    eoc, eoc_expected = runTestPerturbation(C1C0testSpaces, order)
    checkEOC(eoc, eoc_expected)

main()


import numpy as np
from matplotlib import pyplot as plt
import random
import dune.vem
from dune.grid import gridFunction
def reference(h,offset):
    eps = 0.1
    rand = lambda l,r: random.random()*(r-l)+l
    x, y = [], []
    r = random.random()
    if r < 0.2:
        r = 2
    elif r < 0.5:
        r = 3
    elif r < 0.8:
        r = 4
    else:
        r = 5
    if r == 1:
        x += [ h*rand(0.3,0.4) ]
        y += [ h*rand(0.3,0.4) ]
    else:
        for i in range(r):
            x += [ h*rand(i/r+eps,(i+1)/r-eps) ]
            if i%2==0:
                y += [ h*rand(0.2,0.4) ]
            else:
                y += [ h*rand(0.6,0.8) ]
    inner = []
    for i in range(r):
        inner += [ [x[i]+offset[0],y[i]+offset[1]] ]
    return inner

def ncGrid(N):
    h = 1/N
    vert =  []
    for i in range(N+1):
        for j in range(N+1):
            vert += [ [j*h,i*h] ]
    c = 0
    polys = []
    for i in range(N+1):
        for j in range(N+1):
            if i>0 and j>0:
                L = len(vert)
                inner = reference(h, vert[c-N-2])
                vert += inner
                l = len(inner)
                if (i+j)%2 == 0:
                    p1 = [c-N-2,c-N-1,c] + [L+k for k in range(l-1,-1,-1)] + [c-1]
                    p2 = [c-1] + [L+k for k in range(0,l,1)] + [c]
                else:
                    p1 = [c-N-2] + [L+k for k in range(0,l,1)] + [c-N-1]
                    p2 = [c-N-2] + [L+k for k in range(0,l,1)] + [c-N-1,c,c-1]
                polys += [p1]
                polys += [p2]
                # polys += [ [c-N-2,c-N-1,c,L+2,L+1,L,c-1] ]
                # polys += [ [c-1,L,L+1,L+2,c] ]
            c += 1

    cells = {"vertices":np.array(vert), "polygons":polys}
    return cells

def main():
    cells = ncGrid(10)
    dune.vem.writePolygons("concave",cells)
    fig = dune.vem.plotPolygons(cells)
    plt.show()
    """
    grid = dune.vem.polyGrid(cells)
    print(grid.size(0),grid.hierarchicalGrid.agglomerate.size)
    indexSet = grid.indexSet
    @gridFunction(grid, name="cells")
    def polygons(en,x):
        return grid.hierarchicalGrid.agglomerate(indexSet.index(en))
    # polygons.plot()
    """

if __name__ == "__main__":
    main()



import os

path = os.path.join( os.path.dirname(__file__), "tutorial" )
execute  = "cp -RL " + path + " "
execute += "vem_tutorial"
status = os.system(execute)
if status != 0: raise RuntimeError(status)

# remove files generated from CMake
execute  = "rm -rf "
execute += ' '.join("vem_tutorial/{}".format(f) for f in
               ["cmake_install.cmake",
                "CTestTestfile.cmake",
                "Makefile",
                "CMakeFiles",
                "src_dir"
               ])
status = os.system(execute)
if status != 0: raise RuntimeError(status)

print("##################################################################")
print("## Example scripts based on the paper ... is now located in the 'vem_tutorial' folder.")
try:
    import matplotlib
except ImportError:
    print("## Note: the examples requires the installation of 'matplotlib'.")
print("##################################################################")



import matplotlib
matplotlib.rc( 'image', cmap='jet' )
import matplotlib.pyplot as plt
from ufl import *
import dune.ufl, dune.fem, dune.vem
from mixedSolver import MixedSolver # solver for mixed system

Lx,Ly = 1,1.1
x = SpatialCoordinate(triangle)
exact = sin(2*pi*x[0]/Lx)*sin(3*pi*x[1]/Ly)
forcing = -div(grad(exact))

def primal(polyGrid, order):
    spaceU = dune.vem.vemSpace( polyGrid, order=order, testSpaces=[0,order-2,order-2])
    diams  = spaceU.diameters()
    df     = spaceU.interpolate(0,name="solution")
    u,v    = TrialFunction(spaceU), TestFunction(spaceU)
    stiff  = inner(grad(u),grad(v)) * dx - inner(forcing,v) * dx
    dbc    = [dune.ufl.DirichletBC(spaceU,exact,i+1) for i in range(4)]
    scheme = dune.vem.vemScheme([stiff==0,*dbc], gradStabilization=1)
    scheme.solve(target=df)
    return df,diams

def mixed(polyGrid, order):
    spaceU     = dune.vem.bbdgSpace( polyGrid, order=order)
    spaceSigma = dune.vem.curlFreeSpace( polyGrid, order=order)
    diams      = spaceSigma.diameters()
    df         = spaceU.interpolate(0,name="solution")
    sig        = spaceSigma.interpolate([0,0],name="sigma")
    u,v        = TrialFunction(spaceU), TestFunction(spaceU)
    sigma,psi  = TrialFunction(spaceSigma), TestFunction(spaceSigma)
    schemeM = dune.vem.vemScheme(inner(sigma,psi)*dx==0,
                                 massStabilization=1)
    schemeG = dune.fem.operator.galerkin( u*div(psi)*dx )
    schemeD = dune.fem.operator.galerkin( (-div(sigma)-forcing)*v*dx )
    mixedSolver = MixedSolver(schemeM,schemeG,schemeD)
    mixedSolver.solve(target = [df,sig])
    return df,sig,diams

for order in [0,1]:
    for i in range(2,3):
        cells = dune.vem.voronoiCells([[0,0],[Lx,Ly]],2**(2*i+1), lloyd=250)
        polyGrid = dune.vem.polyGrid(cells)
        iset = polyGrid.indexSet
        @dune.fem.function.gridFunction(polyGrid, name="cells", order=0)
        def polygons(en,x):
            return polyGrid.hierarchicalGrid.agglomerate(iset.index(en))
        fig = plt.figure(figsize=(15,15))
        dune.fem.plotting.plotPointData(polygons,figure=fig,
                                        linewidth=2,colorbar=None)
        fig.savefig(f"laplace_{i}.png", dpi=300)

        dfP,diam = primal(polyGrid,order+1)
        dfM,sigma,diam = mixed(polyGrid,order)
        fig = plt.figure(figsize=(25,15))
        dfP.plot(figure=(fig,121),
                        gridLines=None,level=order,colorbar=None)
        dfM.plot(figure=(fig,122),
                        gridLines=None,level=order,colorbar=None)
        fig.savefig(f"laplace_{i}_{order}.png", dpi=300)
